#!/usr/local/bin/ruby

#このファイルの文字コードはUTF-8です。
$KCODE= "u"
require "cgipp"
require "set"
require "nkf"
require "yaml"
require "kana_rome"
load "conf.cgi"


class Item
    
    def initialize(id= nil, params= {})
      @id= id
      @count= 0
      @last_access= Time.now
      assign(params)
    end
    
    def assign(params)
      @title= params["title"]
      @url= params["url"]
      @comment= params["comment"]
    end
    
    def hint
      hints= [
        self.title, self.title_read, self.comment, self.comment_read,
        rome(self.title_read), rome(self.comment_read),
        self.url,
      ]
      return hints.join(" ").downcase()
    end
    
    def add_reading()
      @title_read= reading(@title)
      @comment_read= reading(@comment)
    end
    
    def rome(str)
      return str.split(//).map(){ |c| KanaRome.convert(c) || c }.join("")
    end
    
    def <=>(rhs)
      return self.count <=> rhs.count if @count != rhs.count
      return self.last_access <=> rhs.last_access
    end
    
    def last_access
      return @last_access || Time.at(0)
    end
    
    attr_accessor(:id, :title, :title_read, :url, :comment, :comment_read, :count)
    attr_writer(:last_access)
    
  private
    
    def reading(str)
      return nil if !str
      IO.popen("#{KAKASI} -JH -KH -ieuc -oeuc", "r+") do |f|
        f.print(NKF.nkf("-eW80", str))
        f.close_write()
        return NKF.nkf("-w80E", f.read().chomp())
      end
    end
    
end


class Bookmark
    
    NUM_RECENT_ACCESS= 50
    
    def initialize()
      @items= {}
      @recent_ids= Array.new(NUM_RECENT_ACCESS)
      @next_id= 0
    end
    
    def add(item)
      item.id= @next_id
      @items[@next_id]= item
      @next_id+= 1
      return item
    end
    
    def delete(id)
      @items.delete(id)
    end
    
    def [](id)
      return @items[id]
    end
    
    def access(id)
      dep_id= @recent_ids.shift()
      @recent_ids.push(id)
      @items[dep_id].count-= 1 if @items[dep_id]
      @items[id].count+= 1
      @items[id].last_access= Time.now
    end
    
    def sorted_items
      return @items.values.sort().reverse()
    end
    
    def tags
      tags= Hash.new(0)
      for id, item in @items
        item.comment.gsub(/\[(.*?)\]/) do
          tags[$1]+= item.count
        end
      end
      return tags.keys.sort_by(){ |k| -tags[k] }
    end
    
end


class Main
    
    include(ERB::Util)
    
    def run()
      
      if authorize()
        op= $cgi.param_str("op", "index")
      else
        op= "auth"
      end
      
      script_name= $cgi.script_name
      id= $cgi.param_int("id")
      reload_opener= $cgi.param_bool("reload_opener")
      load()
      
      #操作系。
      case op
        
        when "index", "auth"
        
        when "edit"
          if id
            item= @bookmark[id]
          else
            item= Item.new()
            item.assign(item_params())
              #Bookmarkletで指定された引数をセット。
          end
          if $cgi.has_key?("commit")
            item.assign(item_params())
            item.add_reading()
            @bookmark.add(item) if !id
            save()
            op= "close"
          end
        
        when "delete"
          for str in $cgi.params["id"]
            @bookmark.delete(str.to_i())
          end
          save()
          op= "index"
        
        when "access"
          @bookmark.access(id)
          save()
          op= "none"
        
        else
          raise("No such operation")
        
      end
      
      #表示系。
      case op
        when "index"
          items= @bookmark.sorted_items
          tags= @bookmark.tags
      end
      
      $cgi.run_erb(op+".rhtml", binding())
      
    end
    
  private
    
    def load()
      open(DATA_FILE, "r") do |f|
        f.flock(File::LOCK_SH)
        @bookmark= YAML.load(f)
      end
      @bookmark= Bookmark.new() if !@bookmark
    end
    
    def save()
      open(DATA_FILE, "r+") do |f|
        f.flock(File::LOCK_EX)
        f.truncate(0)
        YAML.dump(@bookmark, f)
      end
    end
    
    def authorize()
      if PASSWORD==""
        return true #パスワード認証OFF。
      elsif $cgi["password"]==PASSWORD || $cgi.cookie["password"][0]==PASSWORD
        $cgi.set_cookie({
          "name" => "password",
          "value" => PASSWORD,
          "expires" => Time.now + 30*24*3600,
        })
        return true
      else
        return false
      end
    end
    
    def item_params()
      params= {}
      for k in ["title", "url", "comment"]
        params[k]= $cgi.unescape_unicode($cgi[k])
      end
      return params
    end
    
    #ERB用。シングルクオート内でのエスケープ。
    def e(str)
      return str.to_s().gsub(/[\'\\]/){ "\\"+$& }
    end
    
end


Main.new().run()
